from typing import Callable, Generator, Literal, TYPE_CHECKING
import html
import markdown
from base64 import b64encode
import re
import io
from . import latex
import pkgutil
from html.parser import HTMLParser
from io import StringIO
from . import svg_tools

HTML_OUTPUT = 0
LATEX_OUTPUT = 1

if TYPE_CHECKING:
    from pandas import DataFrame
    from pandas.io.formats.style import Styler
    import matplotlib.pyplot as plt
    from matplotlib.figure import Figure
    from matplotlib.legend import Legend as Mpl_Legend
    from matplotlib.text import Text as Mpl_Text

    Table = DataFrame | Styler
else:
    try:
        from pandas import DataFrame
    except ImportError:
        DataFrame = None

    try:
        from pandas.io.formats.style import Styler
        Table = DataFrame | Styler
    except ImportError:
        Table = DataFrame
        Styler = None

    try:
        import matplotlib.pyplot as plt
        from matplotlib.figure import Figure
        from matplotlib.legend import Legend as Mpl_Legend
        from matplotlib.text import Text as Mpl_Text
    except ImportError:
        Figure = None


TRenderer = Literal['pandas', 'simple']
FFormat = Literal['svg', 'png', 'pgf']


def _get_pkgutil_string(path: str) -> str:
    data = pkgutil.get_data(__name__, path)
    assert data is not None
    return data.decode()


def _markdown_to_html(text: str) -> str:
    prep_text = re.sub(r'\u00A0', '&nbsp;', text)  # non-breaking space
    html_text = markdown.markdown(prep_text, extensions=['tables', 'fenced_code', 'def_list', 'abbr', 'sane_lists'])
    return html_text


def _drop_indent(text: str, amount: int) -> str:
    """
    Drops a specific number of indentation spaces from a multiline text.

    Args:
        text: The text to drop indentation from
        amount: The number of indentation space characters to drop

    Returns:
        The text with the specified amount of indentation removed
    """
    return ''.join(' ' * amount + line for line in text.splitlines(True))


def _save_figure(fig: Figure, buff: io.BytesIO, figure_format: FFormat, font_family: str | None, scale: float) -> None:
    """
    Saves a matplotlib figure to a file-like object.

    Args:
        fig: The figure to save
        buff: The file-like object to save the figure to
        figure_format: The format to save the figure in (svg, png or pgf)
        font_family: The font family to use for the figure
    """
    def get_all_elements() -> Generator[Mpl_Text, None, None]:
        for ax in fig.get_axes():
            yield ax.title
            yield ax.xaxis.label
            yield ax.yaxis.label
            yield from ax.get_xticklabels() + ax.get_yticklabels()
            legend: Mpl_Legend = ax.get_legend()
            if legend:
                yield from legend.get_texts()

    # Store current figure settings
    old_state = ((e, e.get_fontfamily()) for e in get_all_elements())
    old_size: tuple[float, float] = tuple(fig.get_size_inches())  # type: ignore[unused-ignore]

    # Adjust figure settings
    if font_family:
        for e, _ in old_state:
            e.set_fontfamily(font_family)

    fig.set_size_inches(old_size[0] * scale, old_size[1] * scale, False)

    # Render figure
    backends = {'png': 'AGG', 'svg': 'SVG', 'pgf': 'PGF'}
    assert figure_format in backends, 'Figure format can be pgf (vector), svg (vector) or png (raster)'
    fig.savefig(buff, format=figure_format, backend=backends[figure_format])  # type: ignore[unused-ignore]

    # Reset figure setting
    for e, s in old_state:
        e.set_fontfamily(s)

    fig.set_size_inches(old_size, None, False)


def escape_html(text: str) -> str:
    """
    Escapes special HTML characters in a given string.

    Args:
        text: The text to escape

    Returns:
        Escaped text save for inserting into HTML code
    """
    ret = re.sub(r'\u00A0', '&nbsp;', text)  # non-breaking space
    ret = html.escape(ret)
    return ' '.join(ret.strip().splitlines())


def figure_to_string(fig: Figure,
                     unique_id: str,
                     figure_format: FFormat = 'svg',
                     font_family: str | None = None,
                     scale: float = 1,
                     alt_text: str = '',
                     base64: bool = False) -> str:
    """
    Converts a matplotlib figure to a ascii-string. For png base64 encoding is
    used in general, for svg base64 encoding can be enabled. For base64 encoded
    figures a img-tag is included in the output.

    Args:
        fig: The figure to convert
        figure_format: The format to save the figure in (svg, png or pgf)
        font_family: The font family to use for the figure
        scale: Scaling factor for the figure size
        alt_text: The alt text for the figure
        base64: If the format is svg this determine if the image is encode in base64

    Returns:
        The figure as ascii-string
    """
    assert fig and isinstance(fig, Figure), 'fig parameter must be a matplotlib figure'
    with io.BytesIO() as buff:
        _save_figure(fig, buff, figure_format, font_family, scale)
        buff.seek(0)
        if figure_format == 'pgf':
            i = buff.read(2028).find(b'\\begingroup%')  # skip comments
            buff.seek(max(i, 0))
            return latex.to_ascii(buff.read().decode('utf-8'))

        elif figure_format == 'svg' and not base64:
            i = buff.read(2028).find(b'<svg')  # skip xml and DOCTYPE header
            buff.seek(max(i, 0))
            return svg_tools.update_svg_ids(svg_tools.clean_svg(buff.read().decode('utf-8')), unique_id)

        else:
            image_mime = {"png": "image/png", "svg": "image/svg+xml"}
            assert figure_format in image_mime, 'Unknown image format'
            return '<img alt="%s" src="data:%s;charset=utf-8;base64,%s">' % \
                (escape_html(alt_text),
                 image_mime[figure_format],
                 b64encode(buff.read()).decode('ascii'))  # base64 assures (7-bit) ascii


def latex_to_figure(latex_code: str) -> Figure:
    assert Figure, 'Matplotlib is required for rendering LaTeX expressions for HTML output.'  # type:ignore[truthy-function]
    fig, ax = plt.subplots()
    ax.set_xticks([])
    ax.set_yticks([])
    ax.axis('off')
    text = plt.text(0.5, 0.5, f'${latex_code}$', horizontalalignment='center',
                    verticalalignment='center', transform=ax.transAxes)
    fig.draw_without_rendering()
    bbox = text.get_window_extent()
    fig.set_size_inches(bbox.width / fig.dpi * 1.2, bbox.height / fig.dpi * 1.2)
    return fig


def _fillin_fields(template: str, fields: dict[str, str]) -> str:
    html_out = template
    for variable_name, value in fields.items():
        # Find indentation depths:
        ret = re.search(f"^(.*?)<!---START {variable_name}--->.*?<!---END {variable_name}--->", html_out, flags=re.MULTILINE)
        if ret:
            indent_depths = len(ret.group(1))
            html_out = html_out[:ret.start(0)] + _drop_indent(value, indent_depths) + html_out[ret.end(0):]
    return html_out


def _fillin_reference_names(input_string: str, item_index: dict[str, int]) -> str:
    replacements = [(*m.span(), m.group()) for m in re.finditer(r'(?<=@)\w+:[\w\_\-]+', input_string)]
    ret: list[str] = []
    current_pos = 0
    for start, end, ref in replacements:
        assert ref in item_index, f"Reference {ref} does not exist in the document"
        ret.append(input_string[current_pos:start - 1])
        ret.append(f'<a href="#pyld-ref-{latex.normalize_label_text(ref)}">{item_index[ref]}</a>')
        current_pos = end
    return ''.join(ret) + input_string[current_pos:]


def _check_latex_references(input_string: str, item_index: dict[str, int]) -> str:
    replacements = [m.group() for m in re.finditer(r'(?<=\\ref\{)\w+:[\w\_\\\-]+(?=\})', input_string)]
    escaped_items = set(latex.normalize_label_text(item) for item in item_index)
    for ref in replacements:
        assert ref in escaped_items, f"Reference {ref} does not exist in the document"
    return input_string


def _normalize_text_indent(text: str) -> str:
    text_lines = text.splitlines()
    if len(text_lines) > 1 and not text_lines[0].strip():
        text_lines = text_lines[1:]

    if not text_lines:
        return ''

    if len(text_lines) > 1 and text_lines[0] and text_lines[0][0] != ' ':
        indent_amount = len(text_lines[1]) - len(text_lines[1].lstrip())
    else:
        indent_amount = len(text_lines[0]) - len(text_lines[0].lstrip())

    return '\n'.join(
        [' ' * max(0, len(line) - len(line.strip()) - indent_amount) + line.strip()
         for line in text_lines])


def _create_document_writer() -> 'DocumentWriter':
    new_dwr = DocumentWriter()
    return new_dwr


def inject_to_template(fields_dict: dict[str, str],
                       template_path: str = '',
                       internal_template: str = '',
                       template_string: str = '',
                       ) -> str:
    """
    injects content fields into a template. The placeholder <!--CONTENT-->
    will be replaced by the content. If the placeholder is prefixed with a
    '%' comment character, this character will be replaced as well.

    Args:
        fields_dict: A dictionary with field names as keys and content as values
        template_path: Path to a template file
        internal_template: Path to a internal default template
        template_string: A template string to use directly

    Returns:
        Template with included content
    """
    assert isinstance(fields_dict, dict), 'fields_dict must be a dictionary'
    if template_path:
        with open(template_path, 'r') as f:
            template = f.read()
    elif internal_template:
        template = _get_pkgutil_string(internal_template)
    elif template_string:
        template = template_string
    else:
        raise Exception('No template provided')

    def replace_field(match: re.Match[str]) -> str:
        return fields_dict.get(match.group(1), match.group(0))

    return re.sub(r"(?:\%+\s*)?\<!--(.*?)-->", replace_field, template, 0, re.MULTILINE)


class DocumentWriter():
    """
    A class to create a document for exporting to HTML or LaTeX.
    """
    def __init__(self) -> None:
        """
        Initializes the DocumentWriter instance.
        """
        self._doc: list[list[Callable[[], str]]] = []
        self._fields: dict[str, DocumentWriter] = dict()
        self._base64_svgs: bool = False
        self._figure_format: FFormat = 'svg'
        self._table_renderer: TRenderer = 'simple'
        self._font_family: str | None = None
        self._item_count: dict[str, int] = {}
        self._item_index: dict[str, int] = {}
        self._fig_scale: float = 1

    def _add_item(self, ref_id: str, ref_type: str, caption_prefix: str) -> tuple[str, str]:
        current_index = self._item_count.get(ref_type, 0) + 1
        if not ref_id:
            ref_id = f"auto{current_index}"
        self._item_index[f"{ref_type}:{ref_id}"] = current_index
        self._item_count[ref_type] = current_index
        return caption_prefix.format(current_index), latex.normalize_label_text(f"{ref_type}:{ref_id}")

    def _equation_embedding_reescaping(self, text: str) -> str:
        """
        Convert $$-escaping of LaTeX blocks and inline expressions
        to a HTML-style format: <latex>...</latex>.
        """
        block_pattern = re.compile(
            r'(^|\n)\s*\$\$\s*\n'         # start delimiter on a line on its own
            r'(?P<content>.*?)'           # capture block content non-greedily
            r'\n\s*\$\$\s*(\n|$)',        # end delimiter on a line on its own
            re.DOTALL | re.MULTILINE
        )

        def block_repl(match: re.Match[str]) -> str:
            content = match.group("content").strip()
            latex_label: str = ''

            label_pattern = re.compile(r'^\\label\{([^}]+)\}\s*\n?')
            label_match = label_pattern.match(content)
            if label_match:
                latex_label = label_match.group(1)
                # Remove the label command from the content.
                content = content[label_match.end():].lstrip()

            if latex_label and ':' in latex_label:
                parts = latex_label.split(':')
                ref_type = parts[0]
                ref_id = parts[1]
                caption, reference = self._add_item(ref_id, ref_type, '({})')
                return (f'<latex type="block" reference="{reference}" caption="{caption}">{content}</latex>')
            else:
                return f'<latex type="block">{content}</latex>'

        result = block_pattern.sub(block_repl, text)

        inline_pattern = re.compile(r'\$\$(.+?)\$\$')

        def inline_repl(match: re.Match[str]) -> str:
            content = match.group(1)
            return f'<latex>{content}</latex>'

        return inline_pattern.sub(inline_repl, result)

    def _get_equation_html(self, latex_equation: str, caption: str, reference: str, block: bool = False) -> str:
        fig = latex_to_figure(latex_equation)
        if block:
            fig_str = figure_to_string(fig, reference, self._figure_format, base64=self._base64_svgs)
            ret = ('<div class="equation-container" '
                   f'id="pyld-ref-{reference}">'
                   f'<div class="equation">{fig_str}</div>'
                   f'<div class="equation-number">{caption}</div></div>')
        else:
            ret = '<span class="inline-equation">' + figure_to_string(fig, reference, self._figure_format, base64=self._base64_svgs) + '</span>'

        plt.close(fig)
        return ret

    def _html_post_processing(self, html_code: str) -> str:
        """
        """
        class HTMLPostProcessor(HTMLParser):
            def __init__(self, document_writer: 'DocumentWriter') -> None:
                super().__init__()
                self.modified_html = StringIO()
                self.in_latex: bool = False
                self.eq_caption: str = ''
                self.reference: str = ''
                self.block: bool = False
                self.p_tags: int = 0
                self.dw = document_writer
                self.latex_count = 0
                self.self_closing = False

            def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
                if tag == 'hr':
                    self.modified_html.write(f"<{tag}>")
                    self.self_closing = True
                elif tag == 'latex':
                    self.in_latex = True
                    attr_dict = {k: v if v else '' for k, v in attrs}
                    self.eq_caption = attr_dict.get('caption', '')
                    if 'reference' in attr_dict:
                        self.reference = attr_dict['reference']
                    else:
                        self.latex_count += 1
                        self.reference = f"auto_id_{self.latex_count}"
                    self.block = attr_dict.get('type') == 'block'
                elif not self.in_latex:
                    tag_text = self.get_starttag_text()
                    if tag_text:
                        self.self_closing = tag_text.endswith('/>')
                        self.modified_html.write(tag_text)
                    if tag == 'p':
                        self.p_tags += 1

            def handle_data(self, data: str) -> None:
                if self.in_latex:
                    eq_html = self.dw._get_equation_html(data, self.eq_caption, self.reference, self.block)
                    if self.p_tags > 0 and self.block:
                        # If a block equation (with divs) is inside a p tag: close and reopen it
                        self.modified_html.write(f"</p>{eq_html}<p>")
                    else:
                        self.modified_html.write(eq_html)

                else:
                    self.modified_html.write(data)

            def handle_endtag(self, tag: str) -> None:
                if tag == 'latex':
                    self.in_latex = False
                elif self.self_closing:
                    self.self_closing = False
                else:
                    self.modified_html.write(f"</{tag}>")
                    if tag == 'p' and self.p_tags > 0:
                        self.p_tags -= 1

        parser = HTMLPostProcessor(self)
        parser.feed(html_code)
        return parser.modified_html.getvalue()

    def new_field(self, name: str) -> 'DocumentWriter':
        new_dwr = _create_document_writer()
        self._fields[name] = new_dwr
        return new_dwr

    def add_document(self, doc: 'DocumentWriter') -> None:
        self._doc += doc._doc

    def add_diagram(self, fig: Figure, caption: str = '', ref_id: str = '',
                    prefix_pattern: str = 'Figure {}: ', ref_type: str = 'fig',
                    centered: bool = True) -> None:
        """
        Adds a diagram to the document.

        Args:
            fig: The figure to add (matplotlib figure)
            caption: The caption for the figure
            ref_id: If provided, the figure can be referenced by this string
            prefix_pattern: A custom string for the caption prefix, {} will
                be replaced by the figure number
            ref_type: The type of reference. Each type (e.g., 'fig', 'table')
                has an individual numbering
            centered: Whether to center the figure in LaTeX output
        """

        def render_to_html() -> str:
            caption_prefix, reference = self._add_item(ref_id, ref_type, prefix_pattern)
            return '<div id="pyld-ref-%s" class="figure">%s%s</div>' % (
                reference,
                figure_to_string(fig, reference, self._figure_format, base64=self._base64_svgs, scale=self._fig_scale),
                '<br>' + caption_prefix + escape_html(caption) if caption else '')

        def render_to_latex() -> str:
            _, reference = self._add_item(ref_id, ref_type, prefix_pattern)
            return '\\begin{figure}%s\n%s\n\\caption{%s}\n%s\\end{figure}' % (
                '\n\\centering' if centered else '',
                figure_to_string(fig, reference, 'pgf', self._font_family, scale=self._fig_scale),
                latex.escape_text(caption),
                '\\label{%s}\n' % latex.normalize_label_text(reference) if ref_id else '')

        self._doc.append([render_to_html, render_to_latex])

    def add_table(self, table: Table, caption: str = '', ref_id: str = '',
                  prefix_pattern: str = 'Table {}: ', ref_type: str = 'table', centered: bool = True) -> None:
        """
        Adds a table to the document.

        Args:
            table: The table to add (pandas DataFrame or Styler)
            caption: The caption for the table
            ref_id: If provided, the table can be referenced by this string
            prefix_pattern: A custom string for the caption prefix, {} will
                be replaced by the table number
            ref_type: The type of reference. Each type (e.g., 'fig', 'table')
                has an individual numbering
            centered: Whether to center the table in LaTeX output
        """
        assert Table and isinstance(table, Table), 'Table has to be a pandas DataFrame oder DataFrame Styler'
        styler = table if Styler and isinstance(table, Styler) else getattr(table, 'style', None)  # type: ignore[truthy-function]
        assert Styler and isinstance(styler, Styler), 'Jinja2 package is required for rendering pandas tables'  # type: ignore[truthy-function]

        def render_to_html() -> str:
            caption_prefix, reference = self._add_item(ref_id, ref_type, prefix_pattern)

            html_string = styler.to_html(table_uuid=ref_id, caption=caption_prefix + escape_html(caption))
            return f'<div id="pyld-ref-{reference}">' + re.sub(r'<style.*?>.*?</style>', '', html_string, flags=re.DOTALL) + '</div>'

        def render_to_latex() -> str:
            _, reference = self._add_item(ref_id, ref_type, prefix_pattern)
            if self._table_renderer == 'pandas':
                return styler.to_latex(
                    label=reference,
                    hrules=True,
                    convert_css=True,
                    siunitx=True,
                    caption=latex.escape_text(caption),
                    position_float='centering' if centered else None)
            else:
                return latex.render_pandas_styler_table(styler, caption, reference, centered)

        self._doc.append([render_to_html, render_to_latex])

    def add_text(self, text: str, section_class: str = '') -> None:
        """
        Adds a text paragraph to the document.

        Args:
            text: The text to add
            section_class: The class for the paragraph
        """
        norm_text = _normalize_text_indent(text)

        def render_to_html() -> str:
            html = '<p>' + escape_html(norm_text) + '</p>'
            if section_class:
                return '<div class="' + section_class + '">' + html + '</div>'
            else:
                return html

        def render_to_latex() -> str:
            return latex.from_html(render_to_html())

        self._doc.append([render_to_html, render_to_latex])

    def add_html(self, text: str) -> None:
        """
        Adds HTML formatted text to the document. For the LaTeX
        export only basic HTML for text formatting and tables
        is supported.

        Args:
            text: The HTML to add to the document
        """
        def render_to_html() -> str:
            return text

        def render_to_latex() -> str:
            return latex.from_html(text)

        self._doc.append([render_to_html, render_to_latex])

    def add_h1(self, text: str) -> None:
        """
        Adds a h1 heading to the document.

        Args:
            text: The text of the heading
        """
        def render_to_html() -> str:
            return '<h1>' + escape_html(text) + '</h1>'

        def render_to_latex() -> str:
            return '\\section{' + latex.escape_text(text) + '}\n'

        self._doc.append([render_to_html, render_to_latex])

    def add_h2(self, text: str) -> None:
        """
        Adds a h2 heading to the document.

        Args:
            text: The text of the heading
        """
        def render_to_html() -> str:
            return '<h2>' + escape_html(text) + '</h2>'

        def render_to_latex() -> str:
            return '\\subsection{' + latex.escape_text(text) + '}\n'

        self._doc.append([render_to_html, render_to_latex])

    def add_h3(self, text: str) -> None:
        """
        Adds a h3 heading to the document.

        Args:
            text: The text of the heading
        """
        def render_to_html() -> str:
            return '<h3>' + escape_html(text) + '</h3>'

        def render_to_latex() -> str:
            return '\\subsubsection{' + latex.escape_text(text) + '}\n'

        self._doc.append([render_to_html, render_to_latex])

    def add_equation(self, latex_equation: str, ref_id: str = '', ref_type: str = 'eq') -> None:
        """
        Adds a LaTeX equation to the document.

        Args:
            latex_equation: LaTeX formatted equation
            ref_id: If provided, the equation is displayed with
                a number and can be referenced by the ref_id
        """

        def render_to_html() -> str:
            caption, reference = self._add_item(ref_id, ref_type, '({})')
            return self._get_equation_html(latex_equation, caption, reference, block=True)

        def render_to_latex() -> str:
            _, reference = self._add_item(ref_id, ref_type, '')
            return latex.get_equation_code(latex_equation, reference, block=True)

        self._doc.append([render_to_html, render_to_latex])

    def add_markdown(self, text: str, section_class: str = '') -> None:
        """
        Adds a markdown formatted text to the document.

        Args:
            text: The markdown text to add
            section_class: The class for the text section
        """
        norm_text = _normalize_text_indent(str(text))

        def render_to_html() -> str:
            html = _markdown_to_html(self._equation_embedding_reescaping(norm_text))
            if section_class:
                return '<div class="' + section_class + '">' + html + '</div>'
            else:
                return html

        def render_to_latex() -> str:
            html = _markdown_to_html(
                self._equation_embedding_reescaping(norm_text))
            return latex.from_html(html)

        self._doc.append([render_to_html, render_to_latex])

    def _render_doc(self, doc_type: int) -> str:
        fields = {k: f.to_html() for k, f in self._fields.items()}
        return _fillin_fields(''.join(el[doc_type]() for el in self._doc), fields)

    def to_html(self, figure_format: FFormat = 'svg',
                base64_svgs: bool = False, figure_scale: float = 1) -> str:
        """
        Export the document to HTML. Figures will bew embedded in the HTML code.
        The format can be selected between png in base64, inline svg or svg in base64.

        Args:
            figure_format: The format for embedding the figures in the HTML code (svg or png)
            base64_svgs: Whether to encode svg images in base64

        Returns:
            The HTML code
        """
        self._figure_format = figure_format
        self._base64_svgs = base64_svgs
        self._fig_scale = figure_scale

        return self._html_post_processing(_fillin_reference_names(self._render_doc(HTML_OUTPUT), self._item_index))

    def to_latex(self, font_family: Literal[None, 'serif', 'sans-serif'] = None,
                 table_renderer: TRenderer = 'simple', figure_scale: float = 1) -> str:
        """
        Export the document to LaTeX. Figures will be embedded as pgf graphics.

        Args:
            font_family: Overwrites the front family for figures
            table_renderer: The renderer for tables (simple: renderer with column type
                guessing for text and numbers; pandas: using the internal pandas LaTeX renderer)

        Returns:
            The LaTeX code
        """
        self._font_family = font_family
        assert table_renderer in ['simple', 'pandas'], "table_renderer must be 'simple' or 'pandas'"
        self._table_renderer = table_renderer
        self._fig_scale = figure_scale

        return _check_latex_references(self._render_doc(LATEX_OUTPUT), self._item_index)

    def to_pdf(self, file_path: str,
               font_family: Literal[None, 'serif', 'sans-serif'] = None,
               table_renderer: TRenderer = 'simple',
               latex_template_path: str = '',
               fields_dict: dict[str, str] = {},
               figure_scale: float = 1,
               engine: latex.LatexEngine = 'pdflatex') -> bool:
        """
        Export the document to a PDF file using LaTeX.

        Args:
            file_path: The path to save the PDF file to
            font_family: Overwrites the front family for figures and the template
            table_renderer: The renderer for tables (simple: renderer with column type
                guessing for text and numbers; pandas: using the internal pandas LaTeX renderer)
            latex_template_path: Path to a LaTeX template file. The
                expression <!--CONTENT--> will be replaced by the generated content.
                If no path is provided a default template is used.
            fields_dict: A dictionary with field names as keys and content as values
                replacing the placeholders <!--KEY--> in the template.
            figure_scale: Scaling factor for the figure size
            engine: LaTeX engine (pdflatex, lualatex, xelatex or tectonic)

        Returns:
            True if the PDF file was successfully created
        """
        content = self.to_latex(font_family, table_renderer, figure_scale)
        latex_code = inject_to_template({'CONTENT': content} | fields_dict,
                                        latex_template_path,
                                        internal_template='templates/default_template.tex')

        if font_family == 'sans-serif':
            latex_code = latex.inject_latex_command(latex_code, '\\renewcommand{\\familydefault}{\\sfdefault}')
        success, errors, warnings = latex.compile(latex_code, file_path, engine=engine)

        if not success:
            print('Errors:')
            print('\n'.join(errors))
            print('Warnings:')
            print('\n'.join(warnings))

        return success

    def _repr_html_(self) -> str:
        return self.to_html()

    def __repr__(self) -> str:
        return self.to_html()



import pyladoc.latex


def normalize_latex_code(latex_code: str) -> str:
    return '\n'.join(line.strip() for line in latex_code.splitlines() if line)


def check_only_ascii(latex_code: str) -> bool:
    return all(ord(c) < 128 for c in latex_code)


def test_latex_from_html():
    html_code = """
    <h1>Test</h1>
    <p>This is are Umlautes: Ä,Ö and Ü</p>
    <p>This is a <b>test</b>.</p>
    <p>And this is another <em>test</em>.</p>
    <p>And this is a <strong>third</strong> test.</p>
    <p>And this is a <i>fourth</i> test.</p>
    <p>This is a LaTeX command: \\textbf{test}</p>
    <p>This are typical control characters: {, }, <, >, ", ', &, |, /, \\</p>
    <ul>
        <li>Item 1</li>
        <li>Item 2</li>
    </ul>
    <table>
        <tr>
            <th>Header 1</th>
            <th>Header 2</th>
        </tr>
        <tr>
            <td>Cell 1</td>
            <td>Cell 2</td>
        </tr>
    </table>
    """

    latex_code = pyladoc.latex.from_html(html_code)

    ref_latex_code = r"""
        \section{Test}
        This is are Umlautes: {\"A},{\"O} and {\"U}
        This is a \textbf{test}.
        And this is another \emph{test}.
        And this is a \textbf{third} test.
        And this is a \emph{fourth} test.
        This is a LaTeX command: \textbackslash{}textbf\{test\}
        This are typical control characters: \{, \}, {\textless}, {\textgreater}, ", ', \&, |, /, \textbackslash{}
        \begin{itemize}
        \item Item 1
        \item Item 2
        \end{itemize}
        \begin{tabular}{ll}\toprule
        Header 1 & Header 2 \\
        \midrule
        Cell 1 & Cell 2 \\
        \bottomrule
        \end{tabular}"""

    print(latex_code)

    print('--')

    # print(pyladoc.latex.escape_text(html_code))

    assert check_only_ascii(latex_code), 'Some characters are not ASCII'
    assert normalize_latex_code(ref_latex_code) == normalize_latex_code(latex_code)


def test_latex_from_markdown():
    markdown_code = """
        ## Test1

        | Anz.| Typ      | Beschreibung
        |----:|----------|------------------------------------
        | 12  | BK9050   | Buskoppler
        |  2  | KL1104   | 4 Digitaleingänge
        |  2  | KL2404   | 4 Digitalausgänge (0,5 A)
        |  3  | KL2424   | 4 Digitalausgänge (2 A)
        |  2  | KL4004   | 4 Analogausgänge
        |  1  | KL4002   | 2 Analogausgänge
        | 22  | KL9188   | Potenzialverteilungsklemme
        |  1  | KL9100   | Potenzialeinspeiseklemme
        |  3  | KL3054   | 4 Analogeingänge
        |  5  | KL3214   | PT100 4 Temperatureingänge (3-Leiter)
        |  3  | KL3202   | PT100 2 Temperatureingänge (3-Leiter)
        |  1  | KL2404   | 4 Digitalausgänge
        |  2  | KL9010   | Endklemme

        This is a **test**.

        ## Test2

        | Anz.| Beschreibung
        |----:|------------------------------------
        | 12  | Buskoppler
        |  2  | 4 Digitaleingänge
        |  2  | 4 Digitalausgänge (0,5 A)
        |  3  | 4 Digitalausgänge (2 A)
        |  2  | 4 Analogausgänge
        |  1  | 2 Analogausgänge
    """

    pyla = pyladoc.DocumentWriter()
    pyla.add_markdown(markdown_code)
    latex_code = pyladoc.latex.from_html(pyla.to_html())

    ref_latex_code = r"""
        \subsection{Test1}
        \begin{tabular}{rll}\toprule
        Anz. & Typ & Beschreibung \\
        \midrule
        12 & BK9050 & Buskoppler \\
        2 & KL1104 & 4 Digitaleing{\"a}nge \\
        2 & KL2404 & 4 Digitalausg{\"a}nge (0,5 A) \\
        3 & KL2424 & 4 Digitalausg{\"a}nge (2 A) \\
        2 & KL4004 & 4 Analogausg{\"a}nge \\
        1 & KL4002 & 2 Analogausg{\"a}nge \\
        22 & KL9188 & Potenzialverteilungsklemme \\
        1 & KL9100 & Potenzialeinspeiseklemme \\
        3 & KL3054 & 4 Analogeing{\"a}nge \\
        5 & KL3214 & PT100 4 Temperatureing{\"a}nge (3-Leiter) \\
        3 & KL3202 & PT100 2 Temperatureing{\"a}nge (3-Leiter) \\
        1 & KL2404 & 4 Digitalausg{\"a}nge \\
        2 & KL9010 & Endklemme \\
        \bottomrule
        \end{tabular}
        This is a \textbf{test}.

        \subsection{Test2}
        \begin{tabular}{rl}\toprule
        Anz. & Beschreibung \\
        \midrule
        12 & Buskoppler \\
        2 & 4 Digitaleing{\"a}nge \\
        2 & 4 Digitalausg{\"a}nge (0,5 A) \\
        3 & 4 Digitalausg{\"a}nge (2 A) \\
        2 & 4 Analogausg{\"a}nge \\
        1 & 2 Analogausg{\"a}nge \\
        \bottomrule
        \end{tabular}"""

    print(latex_code)

    assert check_only_ascii(latex_code), 'Some characters are not ASCII'
    assert normalize_latex_code(ref_latex_code) == normalize_latex_code(latex_code)


if __name__ == '__main__':
    test_latex_from_html()
    test_latex_from_markdown()



import pyladoc
import matplotlib.pyplot as plt
import pandas as pd
from . import document_validation
import numpy as np

VALIDATE_HTML_CODE_ONLINE = True
WRITE_RESULT_FILES = True


def mason_saxena_k_mixture(x_h2):
    """
    Calculate the thermal conductivity of a H2/CO2 gas mixture using the Mason and Saxena mixing rule.

    The mixture thermal conductivity is computed as:
        k_mix = Σ_i [ X_i * k_i / (Σ_j X_j * φ_ij) ]
    with:
        φ_ij = 1/√8 * (1 + M_i/M_j)^(-0.5) * [ 1 + ( (k_i/k_j)**0.5 * (M_j/M_i)**0.25 ) ]^2

    Parameters:
        x_h2 (float or array-like): Mole fraction of H2 (from 0 to 1). The CO2 mole fraction is 1 - x_h2.

    Returns:
        float or array-like: Thermal conductivity of the mixture in W/m·K.
    """
    # Pure gas properties (at room temperature, approx.)
    k_h2 = 0.1805   # Thermal conductivity of H2 in W/mK
    k_co2 = 0.0166  # Thermal conductivity of CO2 in W/mK

    M_h2 = 2.016    # Molar mass of H2 in g/mol
    M_co2 = 44.01   # Molar mass of CO2 in g/mol

    # Define the phi_ij function according to Mason and Saxena.
    def phi(k_i, k_j, M_i, M_j):
        return (1 / np.sqrt(8)) * (1 + M_i / M_j) ** (-0.5) * (1 + ((k_i / k_j) ** 0.5 * (M_j / M_i) ** 0.25))**2

    # Compute phi terms for the two species.
    # For i = j the phi terms should be 1.
    phi_h2_h2 = phi(k_h2, k_h2, M_h2, M_h2)       # Should be 1
    phi_h2_co2 = phi(k_h2, k_co2, M_h2, M_co2)
    phi_co2_h2 = phi(k_co2, k_h2, M_co2, M_h2)
    phi_co2_co2 = phi(k_co2, k_co2, M_co2, M_co2)   # Should be 1

    # Ensure we can perform vectorized operations.
    x_h2 = np.array(x_h2)
    x_co2 = 1 - x_h2

    # Use the Mason-Saxena mixing rule:
    # k_mix = X_H2 * k_H2 / (X_H2*φ_H2_H2 + X_CO2*φ_H2_CO2) +
    #         X_CO2 * k_CO2 / (X_CO2*φ_CO2_CO2 + X_H2*φ_CO2_H2)
    k_mix = (x_h2 * k_h2 / (x_h2 * phi_h2_h2 + x_co2 * phi_h2_co2)) + \
            (x_co2 * k_co2 / (x_co2 * phi_co2_co2 + x_h2 * phi_co2_h2))

    return k_mix


def make_document():
    doc = pyladoc.DocumentWriter()

    doc.add_markdown("""
    # Thermal Conductivity of Mixtures

    The determination of the thermal conductivity of gas mixtures is a central aspect of modeling
    transport phenomena, particularly in high-temperature and high-pressure processes. Among the
    most established approaches is the empirical equation introduced by Wassiljewa, which was
    subsequently refined by Mason and Saxena to improve its applicability to multicomponent systems.
    This model offers a reliable means of estimating the thermal conductivity of gas mixtures based
    on the properties of the pure components and their molar interactions.

    The thermal conductivity of a gas mixture, denoted by $$\\lambda_{\\text{mix}}$$, can expressed as
    shown in equation @eq:lambda_mixture.

    $$
    \\label{eq:lambda_mixture}
    \\lambda_{\text{mix}} = \\sum_{i=1}^{n} \\frac{x_i \\lambda_i}{\\sum_{j=1}^{n} x_j \\Phi_{ij}}
    $$

    In this equation, $$x_i$$ represents the molar fraction of component $$i$$ within the mixture,
    while $$\\lambda_i$$ denotes the thermal conductivity of the pure substance $$i$$. The denominator
    contains the interaction parameter $$\\Phi_{ij}$$, which describes the influence of component
    $$j$$ on the transport properties of component $$i$$.

    The interaction parameter $$\\Phi_{ij}$$ is given by the relation shown in equation @eq:interaction_parameter.

    $$
    \\label{eq:interaction_parameter}
    \\Phi_{ij} = \\frac{1}{\\sqrt{8}} \\left(1 + \\frac{M_i}{M_j} \\right)^{-1/2} \\left[ 1 + \\left( \\frac{\\lambda_i}{\\lambda_j} \\right)^{1/2} \\left( \\frac{M_j}{M_i} \\right)^{1/4} \\right]^2
    $$

    Here, $$M_i$$ and $$M_j$$ are the molar masses of the components $$i$$ and $$j$$, respectively.
    Molar masses and thermal conductivity of the pure substances are listed in table @table:gas_probs.
    The structure of this expression illustrates the nonlinear dependence of the interaction term on
    both the molar mass ratio and the square root of the conductivity ratio of the involved species.
    """)

    # Table
    data = {
        "Gas": ["H2", "O2", "N2", "CO2", "CH4", "Ar", "He"],
        "Molar mass in g/mol": ["2.016", "32.00", "28.02", "44.01", "16.04", "39.95", "4.0026"],
        "Thermal conductivity in W/m/K": ["0.1805", "0.0263", "0.0258", "0.0166", "0.0341", "0.0177", "0.1513"]
    }
    df = pd.DataFrame(data)
    doc.add_table(df.style.hide(axis="index"), 'Properties of some gases', 'gas_probs')

    doc.add_markdown("""
    This formulation acknowledges that the transport properties of a gas mixture are not a simple
    linear combination of the individual conductivities. Rather, they are governed by intermolecular
    interactions, which affect the energy exchange and diffusion behavior of each component. These
    interactions are particularly significant at elevated pressures or in cases where the gas components
    exhibit widely differing molecular masses or transport properties.

    The equation proposed by Wassiljewa and refined by Mason and Saxena assumes that binary interactions
    dominate the behavior of the mixture, while higher-order (three-body or more) interactions are
    neglected. It also presumes that the gases approximate ideal behavior, although in practical
    applications, moderate deviations from ideality are tolerated without significant loss of accuracy.
    In figure @fig:mixture the resulting thermal conductivity of an H2/CO2-mixture is shown.""")

    # Figure
    x_h2_values = np.linspace(0, 1, 100)
    k_mixture_values = mason_saxena_k_mixture(x_h2_values)

    fig, ax = plt.subplots()
    ax.set_xlabel("H2 molar fraction / %")
    ax.set_ylabel("Thermal Conductivity / (W/m·K)")
    ax.plot(x_h2_values * 100, k_mixture_values, color='black')

    doc.add_diagram(fig, 'Thermal Conductivity of H2/CO2 mixtures', 'mixture')

    doc.add_markdown("""
    In engineering practice, the accurate determination of $$\\lambda_{\\text{mix}}$$ is essential
    for the prediction of heat transfer in systems such as membrane modules, chemical reactors, and
    combustion chambers. In the context of membrane-based gas separation, for instance, the thermal
    conductivity of the gas mixture influences the local temperature distribution, which in turn affects
    both the permeation behavior and the structural stability of the membrane.

    It is important to note that the calculated mixture conductivity reflects only the gas phase
    behavior. In porous systems such as carbon membranes, additional effects must be considered.
    These include the solid-phase thermal conduction through the membrane matrix, radiative transport
    in pore channels at high temperatures, and transport in the Knudsen regime for narrow pores.
    To account for these complexities, models based on effective medium theory, such as those of
    Maxwell-Eucken or Bruggeman, are frequently employed. These models combine the conductivities of
    individual phases (gas and solid) with geometrical factors that reflect the morphology of the
    porous structure.

    ---

    Expanded by more or less sensible AI jabbering; based on: [doi:10.14279/depositonce-7390](https://doi.org/10.14279/depositonce-7390)
    """)

    return doc


def test_html_render():
    doc = make_document()
    html_code = doc.to_html()

    document_validation.validate_html(html_code, VALIDATE_HTML_CODE_ONLINE)

    if WRITE_RESULT_FILES:
        with open('tests/out/test_html_render1.html', 'w', encoding='utf-8') as f:
            f.write(pyladoc.inject_to_template({'CONTENT': html_code}, internal_template='templates/test_template.html'))


def test_latex_render():
    doc = make_document()

    if WRITE_RESULT_FILES:
        with open('tests/out/test_html_render1.tex', 'w', encoding='utf-8') as f:
            f.write(doc.to_latex())

        assert doc.to_pdf('tests/out/test_latex_render1.pdf', font_family='serif')
    else:
        assert doc.to_pdf('', font_family='serif')  # Write only to temp folder


if __name__ == '__main__':
    test_html_render()
    test_latex_render()



import pyladoc


def test_latex_embedding2():
    test_input = pyladoc._normalize_text_indent("""
    In this equation, $$x_i$$ represents the molar fraction of component $$i$$ within the mixture,
    while $$\\lambda_i$$ denotes the thermal conductivity of the pure substance $$i$$. The denominator
    contains the interaction parameter $$\\Phi_{ij}$$, which describes the influence of component
    $$j$$ on the transport properties of component $$i$$.

    The interaction parameter $$\\Phi_{ij}$$ is given by the relation shown in @eq:ExampleFormula2.

    $$
    \\label{eq:ExampleFormula2}
    \\Phi_{ij} = \\frac{1}{\\sqrt{8}} \\left(1 + \\frac{M_i}{M_j} \\right)^{-1/2} \\left[ 1 + \\left( \\frac{\\lambda_i}{\\lambda_j} \\right)^{1/2} \\left( \\frac{M_j}{M_i} \\right)^{1/4} \\right]^2
    $$
    """)

    expected_output = pyladoc._normalize_text_indent(r"""
    In this equation, <latex>x_i</latex> represents the molar fraction of component <latex>i</latex> within the mixture,
    while <latex>\lambda_i</latex> denotes the thermal conductivity of the pure substance <latex>i</latex>. The denominator
    contains the interaction parameter <latex>\Phi_{ij}</latex>, which describes the influence of component
    <latex>j</latex> on the transport properties of component <latex>i</latex>.

    The interaction parameter <latex>\Phi_{ij}</latex> is given by the relation shown in @eq:ExampleFormula2.<latex type="block" reference="eq:ExampleFormula2" caption="(1)">\Phi_{ij} = \frac{1}{\sqrt{8}} \left(1 + \frac{M_i}{M_j} \right)^{-1/2} \left[ 1 + \left( \frac{\lambda_i}{\lambda_j} \right)^{1/2} \left( \frac{M_j}{M_i} \right)^{1/4} \right]^2</latex>""")

    dummy = pyladoc.DocumentWriter()
    result_string = dummy._equation_embedding_reescaping(test_input)

    print(result_string)
    assert result_string == expected_output


def test_latex_embedding():
    test_input = pyladoc._normalize_text_indent(r"""
    # Test
    $$
    \label{eq:ExampleFormula2}
    \Phi_{ij} = \frac{1}{\sqrt{8}}
    $$
    This $$i$$ is inline LaTeX.
    """)

    expected_output = pyladoc._normalize_text_indent(r"""
    # Test<latex type="block" reference="eq:ExampleFormula2" caption="(1)">\Phi_{ij} = \frac{1}{\sqrt{8}}</latex>This <latex>i</latex> is inline LaTeX.
    """)

    dummy = pyladoc.DocumentWriter()
    result_string = dummy._equation_embedding_reescaping(test_input)

    print(result_string)
    assert result_string == expected_output

    final_html = dummy._html_post_processing(pyladoc._markdown_to_html(result_string))
    print('-- final_html --')
    print(final_html)

    assert '<h1>' in final_html and '<svg ' in final_html and '<div class="equation-number">' in final_html



import pyladoc
import pandas as pd


def test_readme_example():
    doc = pyladoc.DocumentWriter()

    doc.add_markdown("""
        # Example
        This is inline LaTeX: $$\\lambda$$

        This is a LaTeX block with a number:
        $$
        \\label{eq:test1}
        \\lambda_{\text{mix}} = \\sum_{i=1}^{n} \\frac{x_i \\lambda_i}{\\sum_{j=1}^{n} x_j \\Phi_{ij}}
        $$

        This is an example table. The table @table:pandas_example shows some random data.
        """)

    some_data = {
        'Row1': ["Line1", "Line2", "Line3"],
        'Row2': [120, 100, 110],
        'Row3': ['12 g/km', '> 150 g/km', '110 g/km']
    }
    df = pd.DataFrame(some_data)
    doc.add_table(df, 'This is a pandas example table', 'pandas_example')

    html_code = doc.to_html()
    print(html_code)

    assert '<table' in html_code



import re
from re import Match


def update_svg_ids(input_svg: str, unique_id: str) -> str:
    """Add a unique ID part to all svg IDs and update references ti these IDs"""
    id_mapping: dict[str, str] = {}

    def update_ids(match: Match[str]) -> str:
        old_id = match.group(1)
        new_id = f"svg-{unique_id}-{old_id}"
        id_mapping[old_id] = new_id
        return f' id="{new_id}"'

    def update_references(match: Match[str]) -> str:
        old_ref = match.group(1)
        new_ref = id_mapping.get(old_ref, old_ref)
        if match.group(0).startswith('xlink:href'):
            return f'xlink:href="#{new_ref}"'
        else:
            return f'url(#{new_ref})'

    # Update IDs
    svg_string = re.sub(r'\sid="(.*?)"', update_ids, input_svg)

    # Update references to IDs
    svg_string = re.sub(r'url\(#([^\)]+)\)', update_references, svg_string)
    svg_string = re.sub(r'xlink:href="#([^\"]+)"', update_references, svg_string)

    return svg_string


def clean_svg(svg_text: str) -> str:
    # remove all tags not alllowd for inline svg from metadata:
    svg_text = re.sub(r'<metadata>.*?</metadata>', '', svg_text, flags=re.DOTALL)

    # remove illegal path-tags without d attribute:
    return re.sub(r'<path(?![^>]*\sd=)\s.*?/>', '', svg_text, flags=re.DOTALL)



import pyladoc
import matplotlib.pyplot as plt
import pandas as pd
from . import document_validation

VALIDATE_HTML_CODE_ONLINE = False
WRITE_RESULT_FILES = True


def make_document():
    doc = pyladoc.DocumentWriter()

    doc.add_markdown("""
    # Special characters

    ö ä ü Ö Ä Ü ß @ ∆

    π ≈ ± ∆ Σ

    £ ¥ $ €

    Œ

    # Link

    This is a hyperlink: [nonan.net](https://www.nonan.net)

    # Table

    | Anz.| Typ      | Beschreibung
    |----:|----------|------------------------------------
    | 12  | BK9050   | Buskoppler
    |  2  | KL1104   | 4 Digitaleingänge
    |  2  | KL2404   | 4 Digitalausgänge (0,5 A)
    |  3  | KL2424   | 4 Digitalausgänge (2 A)
    |  2  | KL4004   | 4 Analogausgänge
    |  1  | KL4002   | 2 Analogausgänge
    | 22  | KL9188   | Potenzialverteilungsklemme
    |  1  | KL9100   | Potenzialeinspeiseklemme
    |  3  | KL3054   | 4 Analogeingänge
    |  5  | KL3214   | PT100 4 Temperatureingänge (3-Leiter)
    |  3  | KL3202   | PT100 2 Temperatureingänge (3-Leiter)
    |  1  | KL2404   | 4 Digitalausgänge
    |  2  | KL9010   | Endklemme

    ---

    # Equations

    This line represents a reference to the equation @eq:test1.
    """)

    doc.add_equation(r'y = a + b * \sum_{i=0}^{\infty} a_i x^i', 'test1')

    # Figure
    fig, ax = plt.subplots()

    fruits = ['apple', 'blueberry', 'cherry', 'orange']
    counts = [40, 100, 30, 55]
    bar_labels = ['red', 'blue', '_red', 'orange']
    bar_colors = ['tab:red', 'tab:blue', 'tab:red', 'tab:orange']

    ax.bar(fruits, counts, label=bar_labels, color=bar_colors)
    ax.set_ylabel('fruit supply')
    ax.set_title('Fruit supply by kind and color')
    ax.legend(title='Fruit color')

    doc.add_diagram(fig, 'Bar chart with individual bar colors')

    # Table
    mydataset = {
        'Row1': ["Line1", "Line2", "Line3", "Line4", "Line5"],
        'Row2': [120, '95 km/h', 110, '105 km/h', 130],
        'Row3': ['12 g/km', '> 150 g/km', '-110 g/km', '1140 g/km', '\u221213.05 g/km'],
        'Row4': ['5 stars', '4 stars', '5 stars', '4.5 stars', '5 stars'],
        'Row5': [3.5, 7.8, 8.5, 6.9, 4.2],
        'Row6': ['1850 kg', '150 kg', '140 kg', '1600 kg', '17.55 kg'],
        'Row7': ['600 Nm', '250 Nm', '280,8 Nm', '320 Nm', '450 Nm']
    }
    df = pd.DataFrame(mydataset)

    doc.add_table(df.style.hide(axis="index"), 'This is a example table', 'example1')

    return doc


def test_html_render():
    doc = make_document()
    html_code = doc.to_html()

    document_validation.validate_html(html_code, VALIDATE_HTML_CODE_ONLINE)

    if WRITE_RESULT_FILES:
        with open('tests/out/test_html_render2.html', 'w', encoding='utf-8') as f:
            f.write(pyladoc.inject_to_template({'CONTENT': html_code}, internal_template='templates/test_template.html'))


def test_latex_render():
    doc = make_document()

    latex_code = doc.to_latex()

    assert r'\begin{tabular}{lSSSSSS}' in latex_code, "Table format not correct in LaTeX output"

    if WRITE_RESULT_FILES:
        with open('tests/out/test_html_render2.tex', 'w', encoding='utf-8') as f:
            f.write(latex_code)

        assert doc.to_pdf('tests/out/test_latex_render2.pdf', font_family='serif')
    else:
        assert doc.to_pdf('', font_family='serif')  # Write only to temp folder


if __name__ == '__main__':
    test_html_render()
    test_latex_render()



import pyladoc
from . import document_validation

VALIDATE_HTML_CODE_ONLINE = False
WRITE_RESULT_FILES = True


def test_markdown_styling():
    pyla = pyladoc.DocumentWriter()
    pyla.add_markdown(
        """
        Below is an in-depth explanation of the AArch64 (ARM64)
        unconditional branch instruction—often simply called the
        “B” instruction—and how its 26‐bit immediate field (imm26)
        is laid out and later relocated during linking.

        ---

        ## Instruction Layout

        The unconditional branch in AArch64 is encoded in a 32‑bit
        instruction. Its layout is as follows:

        ```
        Bits:  31         26 25                           0
                +-------------+------------------------------+
                |  Opcode     |          imm26               |
                +-------------+------------------------------+
        ```

        - **Opcode (bits 31:26):**
        - For a plain branch (`B`), the opcode is `000101`.
        - For a branch with link (`BL`), which saves the return
        address (i.e., a call), the opcode is `100101`.
        These 6 bits determine the instruction type.

        - **Immediate Field (imm26, bits 25:0):**
        - This 26‑bit field holds a signed immediate value.
        - **Offset Calculation:** At runtime, the processor:
            1. **Shifts** the 26‑bit immediate left by 2 bits.
            (Because instructions are 4-byte aligned,
            the two least-significant bits are always zero.)
            2. **Sign-extends** the resulting 28‑bit value to
            the full register width (typically 64 bits).
            3. **Adds** this value to the program counter
            (PC) to obtain the branch target.

        - **Reach:**
        - With a 26‑bit signed field that’s effectively 28 bits
          after the shift, the branch can cover a range
          of approximately ±128 MB from the current instruction.
        """)

    html_code = pyla.to_html()
    document_validation.validate_html(html_code, check_for=['strong', 'ol', 'li', 'code', 'hr'])

    if WRITE_RESULT_FILES:
        with open('tests/out/test_markdown_style.html', 'w', encoding='utf-8') as f:
            f.write(html_code)


def test_markdown_table():
    pyla = pyladoc.DocumentWriter()
    pyla.add_markdown(
        """
        ## Klemmen

        | Anz.| Typ      | Beschreibung
        |----:|----------|------------------------------------
        | 12  | BK9050   | Buskoppler
        |  2  | KL1104   | 4 Digitaleingänge
        |  2  | KL2404   | 4 Digitalausgänge (0,5 A)
        |  3  | KL2424   | 4 Digitalausgänge (2 A)
        |  2  | KL4004   | 4 Analogausgänge
        |  1  | KL4002   | 2 Analogausgänge
        | 22  | KL9188   | Potenzialverteilungsklemme
        |  1  | KL9100   | Potenzialeinspeiseklemme
        |  3  | KL3054   | 4 Analogeingänge
        |  5  | KL3214   | PT100 4 Temperatureingänge (3-Leiter)
        |  3  | KL3202   | PT100 2 Temperatureingänge (3-Leiter)
        |  1  | KL2404   | 4 Digitalausgänge
        |  2  | KL9010   | Endklemme
        """)

    html_code = pyla.to_html()
    document_validation.validate_html(html_code, check_for=['table'])

    if WRITE_RESULT_FILES:
        with open('tests/out/test_markdown_table.html', 'w', encoding='utf-8') as f:
            f.write(html_code)


def test_markdown_equations():
    pyla = pyladoc.DocumentWriter()
    pyla.add_markdown(
        """
        # Source Equations
        1. $4(3x + 2) - 5(x - 1) = 3x + 14$
        2. $\frac{2y + 5}{4} + \frac{3y - 1}{2} = 5$
        3. $\frac{5}{x + 2} + \frac{2}{x - 2} = 3$
        4. $8(3b - 5) + 4(b + 2) = 60$
        5. $2c^2 - 3c - 5 = 0$
        6. $4(2d - 1) + 5(3d + 2) = 7d + 28$
        7. $q^2 + 6q + 9 = 16$

        # Result Equations
        1. $x = \frac{1}{4}$
        2. $y = \frac{17}{8}$
        3. $z = \frac{7}{3}$
        4. $x = 1$ or $x = -6$
        5. $a = \frac{1}{3}$ or $a = 2$
        6. $x = -\frac{2}{3}$ or $x = 3$
        7. $b = \frac{23}{7}$

        # Step by Step
        1. Distribute: $12x + 8 - 5x + 5 = 3x + 14$
        2. Combine like terms: $7x + 13 = 3x + 14$
        3. Subtract $3x$: $4x + 13 = 14$
        4. Subtract $13$: $4x = 1$
        5. Divide by $4$: $x = \frac{1}{4}$
        """)

    html_code = pyla.to_html()
    document_validation.validate_html(html_code, check_for=['h1'])

    if WRITE_RESULT_FILES:
        with open('tests/out/test_markdown_equations.html', 'w', encoding='utf-8') as f:
            f.write(html_code)


def test_markdown_characters():
    pyla = pyladoc.DocumentWriter()
    pyla.add_markdown(
        """
        # Special caracters

        Umlaute: ÖÄÜ öäü

        Other: ß, €, @, $, %, ~, µ

        Units: m³, cm²

        Controll characters: <, >, ", ', &, |, /, \\

        """)

    html_code = pyla.to_html()
    document_validation.validate_html(html_code, check_for=['h1'])

    if WRITE_RESULT_FILES:
        with open('tests/out/test_markdown_characters.html', 'w', encoding='utf-8') as f:
            f.write(html_code)



from html.parser import HTMLParser
from typing import Generator, Any, Literal, get_args, TYPE_CHECKING
import re
import os
import shutil
import subprocess
import tempfile
from .latex_escaping import unicode_to_latex_dict, latex_escape_dict

if TYPE_CHECKING:
    from pandas.io.formats.style import Styler
else:
    try:
        from pandas.io.formats.style import Styler
    except ImportError:
        Styler = None

LatexEngine = Literal['pdflatex', 'lualatex', 'xelatex', 'tectonic']


def basic_formatter(value: Any) -> str:
    return escape_text(str(value))


def to_ascii(text: str) -> str:
    """
    Replaces/escapes often used unicode characters in LaTeX code or text
    with its LaTeX ascii equivalents.

    Args:
        text: The text to convert.

    Returns:
        The escaped text.
    """
    regex_filter = ('|'.join(unicode_to_latex_dict))

    last_s = 0
    ret: list[str] = []
    for m in re.finditer(regex_filter, text):
        s1, s2 = m.span()
        ret.append(text[last_s:s1])
        ret.append(unicode_to_latex_dict[m.group()])
        last_s = s2
    ret.append(text[last_s:])

    return ''.join(ret)


def normalize_label_text(text: str) -> str:
    """
    Replace any special non-allowed character in the lable text.

    Args:
        text: Input text

    Returns:
        Normalized text
    """
    return re.sub(r"[^a-zA-Z0-9.:]", '-', text)


def escape_text(text: str) -> str:
    """
    Escapes special LaTeX characters and often used unicode characters in a given string.

    Args:
        text: The text to escape

    Returns:
        Escaped text
    """

    latex_translation = latex_escape_dict | unicode_to_latex_dict

    regex_filter = '|'.join(latex_translation)

    last_s = 0
    ret: list[str] = []
    for m in re.finditer(regex_filter, text):
        s1, s2 = m.span()
        ret.append(text[last_s:s1])
        matches = [v for k, v in latex_translation.items() if re.match(k, m.group())]
        if m.group(1):
            ret.append(matches[0].replace(r'\g<1>', normalize_label_text(m.group(1))))
        else:
            ret.append(matches[0])
        last_s = s2
    ret.append(text[last_s:])

    return ''.join(ret)


def get_equation_code(equation: str, reference: str | None, block: bool = False) -> str:
    """
    Converts an equation string to LaTeX code.

    Args:
        equation: The LaTeX equation string.
        reference: The reference type and ID for the equation separated by a ':'.
    """
    if block:
        if reference:
            return '\\begin{equation}\\label{%s}%s\\end{equation}' % (
                normalize_label_text(reference), equation)
        else:
            return '\\[%s\\]' % equation
    else:
        return '\\(%s\\)' % equation


def render_pandas_styler_table(df_style: Styler, caption: str = '', label: str = '', centering: bool = True) -> str:
    """
    Converts a pandas Styler object to LaTeX table.

    Args:
        df_style: The pandas Styler object to convert.
        caption: The caption for the table.
        label: Label for referencing the table.
        centering: Whether to center the table.

    Returns:
        The LaTeX code.
    """
    assert Styler, 'Jinja2 package is required for rendering pandas tables'  # type: ignore[truthy-function]
    assert isinstance(df_style, Styler), 'df_style has to be of type Styler'

    def iter_table(table: dict[str, Any]) -> Generator[str, None, None]:
        yield '\\begin{table}\n'
        if centering:
            yield '\\centering\n'

        # Guess column type
        numeric = re.compile('^[<>\\-\u2212]?\\s*(?:\\d+,?)+(?:\\.\\d+)?(?:\\s\\D.*)?$')
        formats = ['S' if all(
            (numeric.match(line[ci]['display_value'].strip()) for line in table['body'])
        ) else 'l' for ci in range(len(table['body'][0])) if table['body'][0][ci]['is_visible']]

        if caption:
            yield f"\\caption{{{escape_text(caption)}}}\n"
        if label:
            yield f"\\label{{{normalize_label_text(label)}}}\n"
        yield f"\\begin{{tabular}}{{{''.join(formats)}}}\n\\toprule\n"

        for head in table['head']:
            yield (' & '.join(f"\\text{{{escape_text(c['display_value'].strip())}}}"
                              for c in head if c['is_visible']))
            yield ' \\\\\n'

        yield '\\midrule\n'

        for body in table['body']:
            yield (' & '.join(escape_text(c['display_value'].strip())
                              for c in body if c['is_visible']))
            yield ' \\\\\n'

        yield '\\bottomrule\n\\end{tabular}\n\\end{table}'

    str_list = iter_table(df_style._translate(False, False, blank=''))  # type: ignore[attr-defined]

    return ''.join(str_list)


def from_html(html_code: str) -> str:
    """
    Converts HTML code to LaTeX code using HTMLParser.

    Args:
        html_code: The HTML code to convert.

    Returns:
        The LaTeX code.
    """
    html_to_latex = {
        'strong': ('\\textbf{', '}'),
        'b': ('\\textbf{', '}'),
        'em': ('\\emph{', '}'),
        'i': ('\\emph{', '}'),
        'p': ('', '\n\n'),
        'h1': ('\\section{', '}\n'),
        'h2': ('\\subsection{', '}\n'),
        'h3': ('\\subsubsection{', '}\n'),
        'ul': ('\\begin{itemize}\n', '\\end{itemize}\n'),
        'ol': ('\\begin{enumerate}\n', '\\end{enumerate}\n'),
        'li': ('\\item ', '\n')
    }

    class LaTeXHTMLParser(HTMLParser):
        def __init__(self) -> None:
            super().__init__()
            self.latex_code: list[str] = []
            self.header_index: int = -1
            self.column_alignment = ''
            self.midrule_flag = False
            self.header_flag = False
            self.attr_dict: dict[str, str] = {}
            self.equation_flag = False

        def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
            self.attr_dict = {k: v if v else '' for k, v in attrs}
            if tag in html_to_latex:
                prefix, _ = html_to_latex[tag]
                self.latex_code.append(prefix)
            elif tag == 'table':
                self.header_index = len(self.latex_code)
                self.latex_code.append('')  # Placeholder for column header
                self.midrule_flag = False
                self.header_flag = False
            elif tag == 'tr':
                self.column_alignment = ''
            elif tag in ['th', 'td']:
                if 'right' in self.attr_dict.get('style', ''):
                    self.column_alignment += 'r'
                else:
                    self.column_alignment += 'l'
            elif tag == 'a':
                href = self.attr_dict.get('href')
                assert href, 'Link href attribute is missing'
                self.latex_code.append(f"\\href{{{href}}}{{")
            elif tag == 'hr':
                self.latex_code.append("\n\n\\noindent\\rule[0.5ex]{\\linewidth}{1pt}\n\n")
            elif tag == 'latex':
                self.equation_flag = True

        def handle_endtag(self, tag: str) -> None:
            if tag in html_to_latex:
                _, postfix = html_to_latex[tag]
                self.latex_code.append(postfix)
            elif tag == 'table':
                self.latex_code.append("\\bottomrule\n\\end{tabular}\n")
            elif tag == 'tr':
                self.latex_code.pop()  # Remove column separator after last entry
                if self.header_index >= 0:
                    self.latex_code[self.header_index] = f"\\begin{{tabular}}{{{self.column_alignment}}}\\toprule\n"
                    self.header_index = -1
                self.latex_code.append(' \\\\\n')
                if self.header_flag and not self.midrule_flag:
                    self.latex_code.append("\\midrule\n")
                    self.midrule_flag = True
            elif tag == 'th':
                self.latex_code.append(" & ")
                self.header_flag = True
            elif tag == 'td':
                self.latex_code.append(" & ")
            elif tag == 'a':
                self.latex_code.append("}")
            elif tag == 'latex':
                self.equation_flag = False

        def handle_data(self, data: str) -> None:
            if self.equation_flag:
                block = self.attr_dict.get('type') == 'block'
                reference = self.attr_dict.get('reference')
                self.latex_code.append(get_equation_code(data, reference, block))
            elif data.strip():
                self.latex_code.append(escape_text(data))

    parser = LaTeXHTMLParser()
    parser.feed(html_code)
    return ''.join(parser.latex_code)


def compile(latex_code: str, output_file: str = '', encoding: str = 'utf-8', engine: LatexEngine = 'pdflatex') -> tuple[bool, list[str], list[str]]:
    """
    Compiles LaTeX code to a PDF file.

    Args:
        latex_code: The LaTeX code to compile.
        output_file: The output file path.
        encoding: The encoding of the LaTeX code.
        engine: LaTeX engine (pdflatex, lualatex, xelatex or tectonic)

    Returns:
        A tuple with three elements:
        - A boolean indicating whether the compilation was successful.
        - A list of errors.
        - A list of warnings.
    """

    assert engine in get_args(LatexEngine), "engine must be pdflatex, lualatex, xelatex or tectonic"

    with tempfile.TemporaryDirectory() as tmp_path:
        if engine == 'tectonic':
            command = ['tectonic', '--outdir', tmp_path, '-']
        else:
            command = [engine, '--halt-on-error', '--output-directory', tmp_path]

        errors: list[str] = []
        warnings: list[str] = []

        for i in range(1, 4):
            rerun_flag = False
            error_flag = False
            process = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            output, error = process.communicate(input=latex_code.encode(encoding))

            assert not error, 'Compilation error: ' + output.decode(encoding)

            for line in output.decode(encoding).split('\n'):
                if 'Warning' in line:
                    warnings.append(f"Run {i}: " + line)
                    if 'reference' in line:
                        rerun_flag = True
                if line.startswith('!') or line.startswith('*!'):
                    error_flag = True

                if error_flag:
                    errors.append(line)

            if not rerun_flag or errors:
                break

        # Copy pdf file
        file_list = [f for f in os.listdir(tmp_path) if f.lower().endswith('.pdf')]
        if file_list:
            pdf_file = os.path.join(tmp_path, file_list[0])
            if output_file:
                shutil.copyfile(pdf_file, output_file)

    return not errors, errors, warnings


def inject_latex_command(text: str, command: str) -> str:
    lines = text.splitlines()

    last_package_index = -1
    for i, line in enumerate(lines):
        if line.strip().startswith("\\usepackage"):
            last_package_index = i

    if last_package_index != -1:
        lines.insert(last_package_index + 1, f"\n{command}\n")
    else:
        lines.append(f"\n{command}\n")

    return '\n'.join(lines)



import pyladoc


def test_inject_to_template_html():
    template = """
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title><!--TITLE--></title>
    </head>
    <!-- some comment -->
    <body>
        <!--CONTENT-->
    </body>
    </html>
    """

    content = "Hello, World!"
    title = "Test Title"

    result = pyladoc.inject_to_template({'CONTENT': content, 'TITLE': title}, template_string=template)

    print(result)

    assert "Hello, World!" in result
    assert "<!-- some comment -->" in result  # Keep unrelated HTML comments
    assert "<title>Test Title</title>" in result


def test_inject_to_template_latex():
    template = """
\\documentclass[a4paper,12pt]{article}

% Packages
\\usepackage[utf8]{inputenc}
\\usepackage[T1]{fontenc}
\\usepackage{lmodern}  % Load Latin Modern font
\\usepackage{graphicx} % For including images
\\usepackage{amsmath}  % For mathematical symbols
\\usepackage{amssymb}  % For additional symbols
\\usepackage{hyperref} % For hyperlinks
\\usepackage{caption}  % For customizing captions
\\usepackage{geometry} % To set margins
\\usepackage{natbib}   % For citations
\\usepackage{float}    % For fixing figure positions
\\usepackage{siunitx}  % For scientific units
\\usepackage{booktabs} % For professional-looking tables
\\usepackage{pgf} % For using pgf grafics
\\usepackage{textcomp, gensymb} % provides \\degree symbol

\\sisetup{
  table-align-text-post = false
}

% Geometry Settings
\\geometry{margin=1in} % 1-inch margins

% Title and Author Information
\\title{<!--PROJECT-->}
<!--AUTHOR-->
\\date{\\today}

\begin{document}

% Title Page
\\maketitle

% <!--CONTENT-->
\\end{document}
    """

    content = "Hello, World!"
    project_name = "Test Project"
    author_name = "Otto"

    result = pyladoc.inject_to_template(
        {'CONTENT': content, 'PROJECT': project_name, 'AUTHOR': author_name},
        template_string=template)

    print(result)

    assert "\nOtto\n" in result
    assert "\\title{Test Project}\n" in result
    assert "Hello, World!" in result
